"""
Controller class for running commands (list, get, download, upload, update,
                                       verify)
on datafile records.
"""
from __future__ import print_function

import os

from mtclient.models.datafile import DataFile
from mtclient.views import render

from .cli import ModelCliController


class DataFileController(ModelCliController):
    """
    Controller class for running commands (list, get, download, upload, update,
                                           verify))
    on datafile records.
    """
    def __init__(self):
        super(DataFileController, self).__init__()
        self.allowed_commands = [
            "list", "get", "create", "update", "download", "upload", "verify"]
        self.primary_key_arg = "datafile_id"
        self.model = DataFile

    def list(self, args, render_format):
        """
        Display list of datafile records.
        """
        # pylint: disable=no-self-use
        if args.dataset:
            filters = "dataset__id=%s" % args.dataset
        else:
            filters = ""
        if args.directory:
            filters += "&directory=%s" % args.directory
        if args.filename:
            filters += "&filename=%s" % args.filename
        if args.filter:
            filters += "&%s" % args.filter
        datafiles = DataFile.list(
            filters, args.limit, args.offset, args.order_by)
        print(render(datafiles, render_format))

    def get(self, args, render_format):
        """
        Display datafile record.
        """
        # pylint: disable=no-self-use
        datafile = DataFile.objects.get(
            id=args.datafile_id, include_metadata=args.metadata)
        print(render(datafile, render_format))

    def create(self, args, render_format):
        """
        Create datafile record(s) for an existing file or for all files
        within a directory.
        """
        # pylint: disable=no-self-use
        if os.path.isdir(args.path):
            num_created = DataFile.create_datafiles(
                args.dataset_id, args.storagebox, args.dataset_path, args.path)
            print("%s datafiles created." % num_created)
        else:
            datafile = DataFile.create_datafile(
                args.dataset_id, args.storagebox, args.dataset_path, args.path)
            print(render(datafile, render_format))
            print("DataFile created successfully.")

    def download(self, args, _render_format):
        """
        Download datafile.
        """
        # pylint: disable=no-self-use
        DataFile.download(args.datafile_id)

    def upload(self, args, _render_format):
        """
        Upload datafile.
        """
        # pylint: disable=no-self-use
        DataFile.upload(
            args.dataset_id, args.storagebox, args.dataset_path,
            args.file_path)

    def update(self, args, render_format):
        """
        Update datafile record.
        """
        # pylint: disable=no-self-use
        datafile = DataFile.update(args.datafile_id, args.md5sum)
        print(render(datafile, render_format))
        print("DataFile updated successfully.")

    def verify(self, args, _render_format):
        """
        Ask MyTardis to verify a datafile.
        """
        # pylint: disable=no-self-use
        DataFile.verify(args.datafile_id)



"""
Test for command-line yes/no confirmation
"""
import sys
from six import StringIO

import pytest

from mtclient.utils.confirmation import query_yes_no


def test_query_yes_no(capfd):
    """
    Test for command-line yes/no confirmation
    """
    sys_stdin = sys.stdin

    sys.stdin = StringIO("y\n")
    assert query_yes_no("Question?")
    out, _ = capfd.readouterr()
    assert out == "Question? [y/n] "

    sys.stdin = StringIO("n\n")
    assert not query_yes_no("Question?")
    out, _ = capfd.readouterr()
    assert out == "Question? [y/n] "

    sys.stdin = StringIO("\n")
    assert query_yes_no("Question?", default="yes")
    out, _ = capfd.readouterr()
    assert out == "Question? [Y/n] "

    sys.stdin = StringIO("\n")
    assert not query_yes_no("Question?", default="no")
    out, _ = capfd.readouterr()
    assert out == "Question? [y/N] "

    with pytest.raises(ValueError) as err:
        query_yes_no("Question?", default="invalid_default")
        assert str(err) == "invalid default answer: 'invalid_default'"

    sys.stdin = StringIO("invalid\ny\n")
    assert query_yes_no("Question?")
    out, _ = capfd.readouterr()
    assert out == (
        "Question? [y/n] "
        "Please respond with 'yes' or 'no' "
        "(or 'y' or 'n').\n"
        "Question? [y/n] ")

    sys.stdin = sys_stdin



"""
Model class for MyTardis API v1's FacilityResource.
"""

import logging
import requests

from ..conf import config
from .model import Model
from .group import Group

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Facility(Model):
    """
    Model class for MyTardis API v1's FacilityResource.
    """
    def __init__(self, response_dict):
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.name = response_dict['name']
        self.response_dict = response_dict
        self.manager_group = \
            Group(group_json=response_dict['manager_group'])

    def __str__(self):
        """
        Return a string representation of a facility
        """
        return "<%s: %s>" % (type(self).__name__, self.name)

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of facilities.

        :param filters: Filters, e.g. "id=123" or "name=Test Facility"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`Facility` records, encapsulated in a
            `ResultSet` object`.
        """
        from ..utils import add_filters, extend_url
        from .resultset import ResultSet

        url = "%s/api/v1/facility/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(Facility, url, response.json())

    @staticmethod
    def get(**kwargs):
        r"""
        Get facility by ID

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the Facility to retrieve

        :return: A :class:`Facility` record.

        :raises requests.exceptions.HTTPError:
        """
        facility_id = kwargs.get("id")
        if not facility_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for Facility get "
                "at this stage.")
        url = "%s/api/v1/facility/%s/?format=json" % (config.url,
                                                      facility_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return Facility(response.json())



"""
mytardisclient utils
"""
from six.moves import urllib


def human_readable_size_string(num):
    """
    Returns human-readable string.
    """
    num = float(num)
    for unit in ['bytes', 'KB', 'MB', 'GB']:
        if -1024.0 < num < 1024.0:
            return "%3.0f %s" % (num, unit)
        num /= 1024.0
    return "%3.0f %s" % (num, 'TB')


def extend_url(url, limit=None, offset=None, order_by=None):
    """
    Add the limit, offset and order_by to the API request URL
    """
    if limit:
        url += "&limit=%s" % limit
    if offset:
        url += "&offset=%s" % offset
    if order_by:
        url += "&order_by=%s" % order_by
    return url


def add_filters(url, filters):
    """
    Add filters to an API request URL
    """
    if filters:
        filter_components = filters.split('&')
        for filter_component in filter_components:
            field, value = filter_component.split('=')
            url += "&%s=%s" % (field, urllib.parse.quote(value))
    return url


def get_render_format(args):
    """
    Determine how to render the output (ASCII table or JSON)
    depending on whether --json was supplied as a command-line arg
    """
    if hasattr(args, 'json') and args.json:
        return 'json'
    return 'table'



"""
Views for MyTardis dataset records
"""
from __future__ import print_function

import json
from texttable import Texttable


def render_dataset(dataset, render_format):
    """
    Render dataset

    :param dataset: The dataset to be rendered.
    :type dataset: :class:`mtclient.models.dataset.Dataset`
    :param render_format: The format to display the data in ('table' or
        'json').
    """
    if render_format == 'json':
        return render_dataset_as_json(dataset)
    return render_dataset_as_table(dataset)


def render_dataset_as_json(dataset, indent=2, sort_keys=True):
    """
    Returns JSON representation of dataset.

    :param dataset: The dataset to be rendered.
    :type dataset: :class:`mtclient.models.dataset.Dataset`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        dataset.response_dict, indent=indent, sort_keys=sort_keys)


def render_dataset_as_table(dataset):
    """
    Returns ASCII table view of dataset.

    :param dataset: The dataset to be rendered.
    :type dataset: :class:`mtclient.models.dataset.Dataset`
    """
    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Dataset field", "Value"])
    table.add_row(["ID", dataset.id])
    table.add_row(["Experiment(s)", "\n".join(dataset.experiments)])
    table.add_row(["Description", dataset.description])
    table.add_row(["Instrument", dataset.instrument])
    dataset_and_param_sets = table.draw() + "\n"

    for dataset_param_set in dataset.parameter_sets:
        dataset_and_param_sets += "\n"
        table = Texttable(max_width=0)
        table.set_cols_align(["r", 'l', 'l', 'l', 'l', 'l', 'l'])
        table.set_cols_valign(['m', 'm', 'm', 'm', 'm', 'm', 'm'])
        table.header(["DatasetParameter ID", "Schema", "Parameter Name",
                      "String Value", "Numerical Value", "Datetime Value",
                      "Link ID"])
        for dataset_param in dataset_param_set.parameters:
            table.add_row([dataset_param.id,
                           dataset_param.name.schema.name,
                           dataset_param.name.name,
                           dataset_param.string_value or "",
                           dataset_param.numerical_value or '',
                           dataset_param.datetime_value or '',
                           dataset_param.link_id or ''])
        dataset_and_param_sets += table.draw() + "\n"

    return dataset_and_param_sets


def render_datasets(datasets, render_format, display_heading=True):
    """
    Render datasets

    :param datasets: The `ResultSet` of datasets to be rendered.
    :type datasets: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if render_format == 'json':
        return render_datasets_as_json(datasets)
    return render_datasets_as_table(datasets, display_heading)


def render_datasets_as_json(datasets, indent=2, sort_keys=True):
    """
    Returns JSON representation of datasets.

    :param datasets: The result set of datasets to be displayed.
    :type datasets: :class:`mtclient.models.resultset.ResultSet`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        datasets.response_dict, indent=indent, sort_keys=sort_keys)


def render_datasets_as_table(datasets, display_heading=True):
    """
    Returns ASCII table view of datasets.

    :param datasets: The datasets to be rendered.
    :type datasets: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: Setting `display_heading` to True ensures
        that the meta information returned by the query is summarized
        in a 'heading' before displaying the table.  This meta
        information can be used to determine whether the query results
        have been truncated due to pagination.
    """
    heading = "\n" \
        "Model: Dataset\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (datasets.url, datasets.total_count,
           datasets.limit, datasets.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm', 'm'])
    table.header(["Dataset ID", "Experiment(s)", "Description", "Instrument"])
    for dataset in datasets:
        table.add_row([dataset.id, "\n".join(dataset.experiments),
                       dataset.description, dataset.instrument])
    return heading + table.draw() + "\n"



"""
Model class for MyTardis API v1's ExperimentResource.
"""
from __future__ import print_function

import logging

import requests

from ..conf import config
from ..utils import extend_url, add_filters
from .model import Model
from .resultset import ResultSet

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Experiment(Model):
    """
    Model class for MyTardis API v1's ExperimentResource.
    """
    def __init__(self, response_dict, include_metadata=False):
        self.response_dict = response_dict
        self.id = None  # pylint: disable=invalid-name
        self.title = None
        self.description = None
        self.institution_name = None
        for key in self.__dict__:
            if key in response_dict:
                self.__dict__[key] = response_dict[key]
        self.parameter_sets = []
        if include_metadata:
            for exp_param_set_json in response_dict['parameter_sets']:
                self.parameter_sets.append(
                    ExperimentParameterSet(exp_param_set_json))

    def __str__(self):
        """
        Return a string representation of an experiment
        """
        return "<%s: %s>" % (type(self).__name__, self.title)

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by="-created_time"):
        """
        Retrieve a list of experiments.

        :param filters: Filters, e.g. "title=Exp Title"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`Experiment` records, encapsulated in a
            `ResultSet` object.
        """
        url = "%s/api/v1/experiment/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(Experiment, url, response.json())

    @staticmethod
    def get(**kwargs):
        r"""
        Retrieve a single experiment record

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the Experiment to retrieve

        :return: An :class:`Experiment` record.

        :raises requests.exceptions.HTTPError:
        """
        exp_id = kwargs.get("id")
        if not exp_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for Experiment get "
                "at this stage.")
        include_metadata = kwargs.get("include_metadata", False)
        url = "%s/api/v1/experiment/%s/?format=json" % (config.url, exp_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return Experiment(response.json(), include_metadata=include_metadata)

    @staticmethod
    def create(title, description="", institution=None, params_file_json=None):
        """
        Create an experiment record.

        :param title: The title of the experiment.
        :param description: The description of the experiment.
        :param institution: The institution of the experiment.
        :param params_file_json: Path to a JSON file with experiment
            parameters.

        :return: A new :class:`Dataset` record.
        """
        import json
        import os

        new_exp_json = {
            "title": title,
            "description": description,
            "immutable": False
        }
        if institution:
            new_exp_json['institution'] = institution
        if params_file_json:
            assert os.path.exists(params_file_json)
            with open(params_file_json) as params_file:
                new_exp_json['parameter_sets'] = json.load(params_file)
        url = config.url + "/api/v1/experiment/"
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_exp_json))
        response.raise_for_status()
        return Experiment(response.json())

    @staticmethod
    def update(experiment_id, title, description):
        """
        Update an experiment record.
        """
        import json

        updated_fields_json = dict()
        updated_fields_json['title'] = title
        updated_fields_json['description'] = description
        url = "%s/api/v1/experiment/%s/" % \
            (config.url, experiment_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        response.raise_for_status()
        return Experiment(response.json())


class ExperimentParameterSet(object):
    """
    Model class for MyTardis API v1's ExperimentParameterSetResource.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, response_dict):
        from .schema import Schema
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.experiment = response_dict['experiment']
        self.schema = Schema(response_dict['schema'])
        self.parameters = []
        for exp_param_json in response_dict['parameters']:
            self.parameters.append(ExperimentParameter(exp_param_json))

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of experiment parameters.

        :param filters: Filters, e.g. "experiments__id=123"

        :return: A list of :class:`ExperimentParameterSet` records,
            encapsulated in a `ResultSet` object`.
        """
        url = "%s/api/v1/experimentparameterset/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(ExperimentParameterSet, url, response.json())


class ExperimentParameter(object):
    """
    Model class for MyTardis API v1's ExperimentParameterResource.
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, response_dict):
        from .schema import ParameterName
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        pname_id = response_dict['name'].split('/')[-2]
        self.name = ParameterName.objects.get(id=pname_id)
        self.string_value = response_dict['string_value']
        self.numerical_value = response_dict['numerical_value']
        self.datetime_value = response_dict['datetime_value']
        self.link_id = response_dict['link_id']
        self.value = response_dict['value']

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        List experiment parameter records in parameter set.
        """



"""
Controller class for setting up config file.
"""
from __future__ import print_function

import os

from six.moves import input

from mtclient.conf import config


class ConfigController(object):
    """
    Controller class for setting up config file.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, path):
        self.path = path

    def configure(self, args=None):
        """
        Configure MyTardis Client settings.
        """
        if args and hasattr(args, 'key') and args.key:
            print(getattr(config, args.key))
            return

        if os.path.exists(self.path):
            print("A config file already exists at %s" % self.path)
            overwrite = input("Are you sure you want to overwrite it? ")
            if not overwrite.strip().lower().startswith('y'):
                return
            print("")

        config.url = os.environ.get("MYTARDISCLIENT_URL") or \
            input("MyTardis URL? ")
        config.username = os.environ.get("MYTARDISCLIENT_USERNAME") or \
            input("MyTardis Username? ")
        config.apikey = os.environ.get("MYTARDISCLIENT_APIKEY") or \
            input("MyTardis API key? ")
        # Only save settings to config file if at least one of them
        # was supplied by raw input, not by an environment variable:
        if config.url != os.environ.get("MYTARDISCLIENT_URL") or \
                config.username != os.environ.get("MYTARDISCLIENT_USERNAME") or \
                config.apikey != os.environ.get("MYTARDISCLIENT_APIKEY"):
            config.save(self.path)
            print("\nWrote settings to %s" % self.path)



"""
This module contains the :class:`ResultSet` class, an abstraction to represent
the JSON returned by the MyTardis API, particularly for queries which return
multiple records and could be subject to pagination.
"""


class ResultSet(object):
    """
    Abstraction to represent JSON returned by MyTardis API
    which includes a list of records and some meta information
    e.g. whether there are additional pages of records to
    retrieve.
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, model, url, response_dict):
        """
        Each record in the result set can be
        represented as an object of class model
        """
        self.model = model
        self.url = url
        self.response_dict = response_dict
        self.total_count = self.response_dict['meta']['total_count']
        self.limit = self.response_dict['meta']['limit']
        self.offset = self.response_dict['meta']['offset']
        self.next = self.response_dict['meta']['next']
        self._objects = None

    def __repr__(self):
        """
        String representation
        """
        if not self._objects:
            self._objects = []
            for index in range(len(self.response_dict['objects'])):
                self._objects.append(self.model(self.response_dict['objects'][index]))
        return "<ResultSet [%s]>" % ", ".join(str(obj) for obj in self._objects)

    def __len__(self):
        """
        Return number of records in ResultSet
        """
        return len(self.response_dict['objects'])

    def __bool__(self):
        """
        Return True if there's a list one object in the result set
        """
        return len(self) > 0

    def __nonzero__(self):
        """
        For Python 2.x
        """
        return self.__bool__()

    def __getitem__(self, key):
        """
        Get a record from the query set.
        """
        if 'include_metadata' in self.model.__init__.__code__.co_varnames:
            return self.model(self.response_dict['objects'][key],
                              include_metadata=False)
        return self.model(self.response_dict['objects'][key])

    def __iter__(self):
        """
        Return the ResultSet's iterator object, which is itself.
        """
        kwargs = {}
        if 'include_metadata' in self.model.__init__.__code__.co_varnames:
            kwargs['include_metadata'] = False
        for index in range(len(self.response_dict['objects'])):
            args = [self.response_dict['objects'][index]]
            yield self.model(*args, **kwargs)

    @staticmethod
    def empty(model):
        """
        Return an empty result set
        """
        response_dict = {
            "meta": {
                "limit": 20,
                "next": None,
                "offset": 0,
                "previous": None,
                "total_count": 0
            },
            "objects": [
            ]
        }
        return ResultSet(model, None, response_dict)



"""
Views for MyTardis facility records
"""
from __future__ import print_function

import json
from texttable import Texttable


def render_facility(facility, render_format, display_heading=True):
    """
    Render facility

    :param facility: The facility to be rendered.
    :type facility: :class:`mtclient.models.facility.Facility`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for
        an `ApiEndpoints` set, setting `display_heading` to True
        ensures that a heading is displayed before the results table.
        The heading includes the URL resolved to perform the query.
    """
    if render_format == 'json':
        return render_facility_as_json(facility)
    return render_facility_as_table(facility, display_heading)


def render_facility_as_json(facility, indent=2, sort_keys=True):
    """
    Returns JSON representation of facility.

    :param facility: The facility to be rendered.
    :type facility: :class:`mtclient.models.facility.Facility`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        facility.response_dict, indent=indent, sort_keys=sort_keys)


def render_facility_as_table(facility, display_heading=True):
    """
    Returns ASCII table view of facility.

    :param facility: The facility to be rendered.
    :type facility: :class:`mtclient.models.facility.Facility`
    :param display_heading: When using the 'table' render format for
        an `ApiEndpoints` set, setting `display_heading` to True
        ensures that a heading is displayed before the results table.
        The heading includes the URL resolved to perform the query.
    """
    heading = "\nModel: Facility\n\n" if display_heading else ""

    table = Texttable()
    table.set_cols_align(['l', 'l'])
    table.set_cols_valign(['m', 'm'])
    table.header(["Facility field", "Value"])
    table.add_row(["ID", facility.id])
    table.add_row(["Name", facility.name])
    table.add_row(["Manager Group", facility.manager_group])
    return heading + table.draw() + "\n"


def render_facilities(facilities, render_format, display_heading=True):
    """
    Render facilities

    :param facilities: The `ResultSet` of facilities to be rendered.
    :type facilities: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if render_format == 'json':
        return render_facilities_as_json(facilities)
    return render_facilities_as_table(facilities, display_heading)


def render_facilities_as_json(facilities, indent=2, sort_keys=True):
    """
    Returns JSON representation of facilities.

    :param facilities: The result set of facilities to be displayed.
    :type facilities: :class:`mtclient.models.resultset.ResultSet`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        facilities.response_dict, indent=indent, sort_keys=sort_keys)


def render_facilities_as_table(facilities, display_heading=True):
    """
    Returns ASCII table view of facilities.

    :param facilities: The facilities to be rendered.
    :type facilities: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: Setting `display_heading` to True ensures
        that the meta information returned by the query is summarized
        in a 'heading' before displaying the table.  This meta
        information can be used to determine whether the query results
        have been truncated due to pagination.
    """
    heading = "\n" \
        "Model: Facility\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (facilities.url, facilities.total_count,
           facilities.limit, facilities.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Name", "Manager Group"])
    for facility in facilities:
        table.add_row([facility.id, facility.name, facility.manager_group])
    return heading + table.draw() + "\n"



"""
test_dataset.py

Tests for functionality to query the Dataset model via
the MyTardis REST API's dataset resource
"""
import json
import requests_mock

from mtclient.conf import config
from mtclient.models.dataset import Dataset


def test_dataset_list():
    """
    Test listing dataset records exposed in a MyTardis REST API
    """
    mock_dataset_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [{
            "created_time": None,
            "description": "dataset description",
            "directory": None,
            "experiments": [
                "/api/v1/experiment/1/"
            ],
            "id": 1,
            "immutable": False,
            "instrument": {
                "created_time": None,
                "facility": {
                    "created_time": None,
                    "id": 1,
                    "manager_group": {
                        "id": 1,
                        "name": "test-facility-managers",
                        "resource_uri": "/api/v1/group/1/"
                    },
                    "modified_time": None,
                    "name": "Test Facility",
                    "resource_uri": "/api/v1/facility/1/"
                },
                "id": 1,
                "modified_time": None,
                "name": "Test Instrument",
                "resource_uri": "/api/v1/instrument/1/"
            },
            "modified_time": None,
            "parameter_sets": [],
            "resource_uri": "/api/v1/dataset/1/"
        }]
    }
    mock_list_response = json.dumps(mock_dataset_list)
    with requests_mock.Mocker() as mocker:
        list_datasets_url = "%s/api/v1/dataset/?format=json&experiments__id=1" % config.url
        mocker.get(list_datasets_url, text=mock_list_response)
        datasets = Dataset.list(filters="experiments__id=1")
        assert datasets.response_dict == mock_dataset_list


def test_dataset_get():
    """
    Test getting a dataset record by ID
    """
    mock_dataset = {
        "created_time": None,
        "description": "dataset description",
        "directory": None,
        "experiments": [
            "/api/v1/experiment/1/"
        ],
        "id": 1,
        "immutable": False,
        "instrument": {
            "created_time": None,
            "facility": {
                "created_time": None,
                "id": 1,
                "manager_group": {
                    "id": 1,
                    "name": "test-facility-managers",
                    "resource_uri": "/api/v1/group/1/"
                },
                "modified_time": None,
                "name": "Test Facility",
                "resource_uri": "/api/v1/facility/1/"
            },
            "id": 1,
            "modified_time": None,
            "name": "Test Instrument",
            "resource_uri": "/api/v1/instrument/1/"
        },
        "modified_time": None,
        "parameter_sets": [],
        "resource_uri": "/api/v1/dataset/1/"
    }
    mock_get_response = json.dumps(mock_dataset)
    with requests_mock.Mocker() as mocker:
        get_dataset_url = "%s/api/v1/dataset/1/?format=json" % config.url
        mocker.get(get_dataset_url, text=mock_get_response)
        dataset = Dataset.objects.get(id=1)
        assert dataset.response_dict == mock_dataset


def test_dataset_create():
    """
    Test creating a dataset record
    """
    mock_dataset = {
        "created_time": None,
        "description": "dataset description",
        "directory": None,
        "experiments": [
            "/api/v1/experiment/1/"
        ],
        "id": 1,
        "immutable": False,
        "instrument": {
            "created_time": None,
            "facility": {
                "created_time": None,
                "id": 1,
                "manager_group": {
                    "id": 1,
                    "name": "test-facility-managers",
                    "resource_uri": "/api/v1/group/1/"
                },
                "modified_time": None,
                "name": "Test Facility",
                "resource_uri": "/api/v1/facility/1/"
            },
            "id": 1,
            "modified_time": None,
            "name": "Test Instrument",
            "resource_uri": "/api/v1/instrument/1/"
        },
        "modified_time": None,
        "parameter_sets": [],
        "resource_uri": "/api/v1/dataset/1/"
    }
    mock_post_response = json.dumps(mock_dataset)
    with requests_mock.Mocker() as mocker:
        post_dataset_url = "%s/api/v1/dataset/" % config.url
        mocker.post(post_dataset_url, text=mock_post_response)
        dataset = Dataset.objects.create(
            experiment_id=1, instrument_id=1,
            description="dataset description")
        assert dataset.response_dict == mock_dataset



"""
Views for MyTardis records
"""
from __future__ import print_function

import json
from texttable import Texttable

from ..models.api import ApiEndpoints, ApiSchema
from ..models.facility import Facility
from ..models.instrument import Instrument
from ..models.experiment import Experiment
from ..models.dataset import Dataset
from ..models.datafile import DataFile
from ..models.storagebox import StorageBox
from ..models.schema import Schema
from ..models.resultset import ResultSet

from .api import render_api_endpoints, render_api_schema
from .facility import render_facility, render_facilities
from .instrument import render_instrument, render_instruments
from .experiment import render_experiment, render_experiments
from .dataset import render_dataset, render_datasets
from .datafile import render_datafile, render_datafiles
from .storagebox import render_storage_box, render_storage_boxes
from .schema import render_schema, render_schemas


def render(data, render_format='table', display_heading=True):
    """
    Generic render function.

    Calls a more specific render function depending on the data type
    to display (render) the data in the desired format.

    :param data: The data to be displayed.  An instance of a model class
        (e.g.  :class:`mtclient.models.dataset.Dataset`) or an instance of
        :class:`mtclient.models.resultset.ResultSet`
        or an instance of :class:`mtclient.models.api.ApiEndpoints`.
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if data.__class__ == ResultSet:
        return render_result_set(data, render_format, display_heading)
    if data.__class__ == ApiEndpoints:
        return render_api_endpoints(data, render_format, display_heading)
    return render_single_record(data, render_format)


def render_single_record(data, render_format):
    """
    Render single record.

    Calls a more specific render function depending on the data type
    to display (render) the data in the desired format.

    :param data: The data to be displayed.  An instance of a model class
        (e.g.  :class:`mtclient.models.dataset.Dataset`).
    :param render_format: The format to display the data in ('table' or
        'json').
    """
    if not data:
        return ""
    renderers = dict(
        ApiSchema=render_api_schema,
        Facility=render_facility,
        Instrument=render_instrument,
        Experiment=render_experiment,
        Dataset=render_dataset,
        DataFile=render_datafile,
        StorageBox=render_storage_box,
        Schema=render_schema)
    data_type = data.__class__.__name__
    if data_type not in renderers:
        raise NotImplementedError("Unexpected data type: %s" % data_type)
    return renderers[data_type](data, render_format)


def render_result_set(result_set, render_format, display_heading=True):
    """
    Render result set.

    Calls a more specific render function depending on the type of data
    stored within the `ResultSet` to display (render) the data in the
    desired format.

    :param result_set: The result set to be rendered.
    :type result_set: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    renderers = dict(
        Facility=render_facilities,
        Instrument=render_instruments,
        Experiment=render_experiments,
        Dataset=render_datasets,
        DataFile=render_datafiles,
        StorageBox=render_storage_boxes,
        Schema=render_schemas)
    data_type = result_set.model.__name__
    if data_type not in renderers:
        raise NotImplementedError("Unexpected data type: %s" % data_type)
    return renderers[data_type](result_set, render_format, display_heading)



"""
Model class for the configuration,
usually stored in ~/.config/mytardisclient/mytardisclient.cfg
"""
import os
import json

from configparser import ConfigParser
from six.moves import urllib

from ..utils.exceptions import InvalidConfig

DEFAULT_CONFIG_PATH = os.path.join(os.path.expanduser('~'), '.config',
                                   'mytardisclient', 'mytardisclient.cfg')
DATASETS_PATH_PREFIX = os.path.join(os.path.expanduser('~'), '.config',
                                    'mytardisclient', 'servers')
LOGFILE_PATH = os.path.join(os.path.expanduser('~'), '.mytardisclient.log')
LOGGING_CONFIG_PATH = os.path.join(os.path.expanduser('~'), '.config',
                                   'mytardisclient', 'logging.cfg')
DEFAULT_LOGGING_CONF = """\
[loggers]
keys=root

[handlers]
keys=fileHandler,consoleHandler

[formatters]
keys=simpleFormatter

[logger_root]
level=DEBUG
handlers=fileHandler,consoleHandler

[handler_fileHandler]
class=FileHandler
level=DEBUG
formatter=simpleFormatter
args=('%(file_path)s', 'a')

[handler_consoleHandler]
class=StreamHandler
level=WARNING
formatter=simpleFormatter
args=(sys.stdout,)

[formatter_simpleFormatter]
format=%%(asctime)s - %%(name)s - %%(levelname)s - %%(message)s
datefmt=
""" % {'file_path': LOGFILE_PATH}


class Config(object):
    """
    Model class for the minimal MyTardis server configuration
    (MyTardis URL, username and API key),
    usually stored in ~/.config/mytardisclient/mytardisclient.cfg
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, path=DEFAULT_CONFIG_PATH):
        #: The config file's location.
        #: Default: ~/.config/mytardisclient/mytardisclient.cfg
        self.path = path

        #: The logging config path.
        #: Default: ~/.config/mytardisclient/logging.cfg
        self.logging_config_path = LOGGING_CONFIG_PATH
        if not os.path.exists(os.path.dirname(self.logging_config_path)):
            os.makedirs(os.path.dirname(self.logging_config_path))
        if not os.path.exists(self.logging_config_path):
            with open(self.logging_config_path, 'w') as logging_config:
                logging_config.write(DEFAULT_LOGGING_CONF)
        self.logfile_path = LOGFILE_PATH

        #: The MyTardis URL, e.g. 'http://mytardisdemo.erc.monash.edu.au'
        self.url = ""

        #: The MyTardis username, e.g. 'demofacility"
        self.username = ""

        #: The MyTardis API key, e.g. '644be179cc6773c30fc471bad61b50c90897146c'
        self.apikey = ""

        if path:
            self.load()

    def __unicode__(self):
        attrs = dict(path=self.path,
                     logfile_path=self.logfile_path,
                     logging_config_path=self.logging_config_path,
                     url=self.url,
                     username=self.username,
                     apikey=self.apikey,
                     datasets_path=self.datasets_path)
        return json.dumps(attrs, indent=2)

    def __str__(self):
        return self.__unicode__()

    def __repr__(self):
        return self.__unicode__()

    @property
    def hostname(self):
        """
        Determine the MyTardis hostname from the MyTardis URL
        """
        parsed_url = urllib.parse.urlparse(self.url)
        return parsed_url.netloc

    @property
    def datasets_path(self):
        """
        Location to create symlinks to dataset folders.
        Default: ~/.config/mytardisclient/servers/[mytardis_hostname]/
        """
        datasets_path = os.path.join(DATASETS_PATH_PREFIX, self.hostname)
        if not os.path.exists(datasets_path):
            os.makedirs(datasets_path)
        return datasets_path

    def load(self, path=None):
        """
        Sets some default values for settings fields, then loads a config
        file.

        :param path: The path to the config file, usually
            ~/.config/mytardisclient/mytardisclient.cfg
        """
        self.url = os.environ.get("MYTARDISCLIENT_URL", "")
        self.username = os.environ.get("MYTARDISCLIENT_USERNAME", "")
        self.apikey = os.environ.get("MYTARDISCLIENT_APIKEY", "")

        if path:
            self.path = path
        else:
            path = self.path

        if path is not None and os.path.exists(path):
            config_parser = ConfigParser()
            config_parser.read(path)
            section = "mytardisclient"
            fields = ["url", "username", "apikey"]

            for field in fields:
                if config_parser.has_option(section, field):
                    self.__dict__[field] = \
                        config_parser.get(section, field)

    @property
    def default_headers(self):
        """
        Default headers to use for API queries
        (including API key authorization).
        """
        headers = {
            "Content-Type": "application/json",
            "Accept": "application/json"
        }
        if self.username and self.apikey:
            headers["Authorization"] = \
                "ApiKey %s:%s" % (self.username, self.apikey)
        return headers

    def validate(self):
        """
        Ensure that the config contains a non-empty username,
        API key and MyTardis URL.
        """
        if self.username == "":
            raise InvalidConfig("MyTardis username is missing from config.")
        if self.apikey == "":
            raise InvalidConfig("MyTardis API key is missing from config.")
        if self.url == "":
            raise InvalidConfig("MyTardis URL is missing from config.")
        parsed_url = urllib.parse.urlparse(self.url)
        if parsed_url.scheme not in ('http', 'https') or \
                parsed_url.netloc == '':
            raise InvalidConfig("Invalid MyTardis URL found in config: %s"
                                % self.url)
        return True

    def save(self, path=None):
        """
        Saves the configuration to disk.

        :param path: The path to save to, usually
            ~/.config/mytardisclient/mytardisclient.cfg
        """
        if path:
            self.path = path
        else:
            path = self.path
        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path))

        config_parser = ConfigParser()
        with open(self.path, 'w') as config_file:
            config_parser.add_section("mytardisclient")
            fields = ["url", "username", "apikey"]
            for field in fields:
                config_parser.set("mytardisclient", field, self.__dict__[field])
            config_parser.write(config_file)



"""
Custom exceptions to raise within MyTardis Client.
"""


class DuplicateKey(Exception):
    """
    Duplicate key exception.
    """


class SshException(Exception):
    """
    SSH exception.
    """
    def __init__(self, message, returncode=None):
        super(SshException, self).__init__(message)
        self.returncode = returncode


class InvalidConfig(Exception):
    """
    Invalid config.
    """


class MissingConfig(Exception):
    """
    Missing config.
    """



"""
test_schema.py

Tests for functionality to query the Schema model via
the MyTardis REST API's facility resource
"""
import json

import pytest
import requests_mock

from mtclient.conf import config
from mtclient.models.schema import Schema, ParameterName


def test_schema_list():
    """
    Test listing schema records exposed in a MyTardis REST API
    """
    mock_schema_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "hidden": False,
                "id": 1,
                "immutable": True,
                "name": "Schema Name",
                "namespace": "http://schema/namespace",
                "resource_uri": "/api/v1/schema/1/",
                "subtype": "",
                "type": 1
            }
        ]
    }
    mock_list_response = json.dumps(mock_schema_list)
    with requests_mock.Mocker() as mocker:
        list_schemas_url = "%s/api/v1/schema/?format=json" % config.url
        mocker.get(list_schemas_url, text=mock_list_response)
        schemas = Schema.objects.all()
        schemas._execute_query()
        assert schemas._result_set.response_dict == mock_schema_list


def test_schema_get():
    """
    Test getting a schema record by ID
    """
    mock_schema = {
        "hidden": False,
        "id": 1,
        "immutable": True,
        "name": "Schema Name",
        "namespace": "http://schema/namespace",
        "resource_uri": "/api/v1/schema/1/",
        "subtype": "",
        "type": 1
    }
    mock_get_response = json.dumps(mock_schema)
    with requests_mock.Mocker() as mocker:
        get_schema_url = "%s/api/v1/schema/1/?format=json" % config.url
        mocker.get(get_schema_url, text=mock_get_response)
        schema = Schema.objects.get(id=1)
        assert schema.response_dict == mock_schema
        assert str(schema) == "<Schema: Schema Name>"
        with pytest.raises(NotImplementedError) as err:
            Schema.objects.get(invalid_key="value")
            assert str(err) == (
                "Only the id keyword argument is supported for Schema get "
                "at this stage.")


def test_param_names_list():
    """
    Test getting a list of parameter names from a schema ID
    """
    mock_pnames_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "choices": "",
                "comparison_type": 1,
                "data_type": 2,
                "full_name": "Parameter Name",
                "id": 1,
                "immutable": True,
                "is_searchable": False,
                "name": "param name",
                "order": 1,
                "resource_uri": "/api/v1/parametername/1/",
                "schema": "/api/v1/schema/1/",
                "units": ""
            }
        ]
    }
    mock_list_response = json.dumps(mock_pnames_list)
    with requests_mock.Mocker() as mocker:
        list_pnames_url = "%s/api/v1/parametername/?format=json" % config.url
        mocker.get(list_pnames_url, text=mock_list_response)
        param_names = ParameterName.list(filters="schema__id=1")
        assert param_names.response_dict == mock_pnames_list


def test_pname_get():
    """
    Test getting a parameter name record by ID
    """
    mock_schema = {
        "hidden": False,
        "id": 1,
        "immutable": True,
        "name": "Schema Name",
        "namespace": "http://schema/namespace",
        "resource_uri": "/api/v1/schema/1/",
        "subtype": "",
        "type": 1
    }
    mock_schema_response = json.dumps(mock_schema)
    mock_pname = {
        "choices": "",
        "comparison_type": 1,
        "data_type": 2,
        "full_name": "Parameter Name",
        "id": 1,
        "immutable": True,
        "is_searchable": False,
        "name": "param name",
        "order": 1,
        "resource_uri": "/api/v1/parametername/1/",
        "schema": "/api/v1/schema/1/",
        "units": ""
    }
    mock_pname_response = json.dumps(mock_pname)
    with requests_mock.Mocker() as mocker:
        get_schema_url = "%s/api/v1/schema/1/?format=json" % config.url
        mocker.get(get_schema_url, text=mock_schema_response)
        get_pname_url = "%s/api/v1/parametername/1/?format=json" % config.url
        mocker.get(get_pname_url, text=mock_pname_response)
        pname = ParameterName.objects.get(id=1)
        assert pname.response_dict == mock_pname
        assert str(pname) == "<ParameterName: Parameter Name>"
        with pytest.raises(NotImplementedError) as err:
            ParameterName.objects.get(invalid_key="value")
            assert str(err) == (
                "Only the id keyword argument is supported for ParameterName get "
                "at this stage.")



"""
test_experiment.py

Tests for functionality to query the Experiment model via
the MyTardis REST API's experiment resource
"""
import json
import requests_mock

from mtclient.conf import config
from mtclient.models.experiment import Experiment


def test_experiment_list():
    """
    Test listing experiment records exposed in a MyTardis REST API
    """
    mock_experiment_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [{
            "id": 1,
            "title": "Test Experiment",
            "description": "exp description",
            "authors": [
            ],
            "created_by": "/api/v1/user/1/",
            "created_time": "2017-08-03T12:59:48.600082",
            "institution_name": "Monash University",
            "owner_ids": [
                1,
            ],
            "parameter_sets": [
            ],
            "public_access": 1,
            "update_time": "2017-08-03T12:59:48.600106",
            "resource_uri": "/api/v1/experiment/1/",
        }]
    }
    mock_list_response = json.dumps(mock_experiment_list)
    with requests_mock.Mocker() as mocker:
        list_exps_url = "%s/api/v1/experiment/?format=json" % config.url
        mocker.get(list_exps_url, text=mock_list_response)
        experiments = Experiment.list()
        assert experiments.response_dict == mock_experiment_list


def test_experiment_get():
    """
    Test getting a experiment record by ID
    """
    mock_experiment = {
        "id": 1,
        "title": "Test Experiment",
        "description": "exp description",
        "authors": [
        ],
        "created_by": "/api/v1/user/1/",
        "created_time": "2017-08-03T12:59:48.600082",
        "institution_name": "Monash University",
        "owner_ids": [
            1,
        ],
        "parameter_sets": [
        ],
        "public_access": 1,
        "update_time": "2017-08-03T12:59:48.600106",
        "resource_uri": "/api/v1/experiment/1/",
    }
    mock_get_response = json.dumps(mock_experiment)
    with requests_mock.Mocker() as mocker:
        get_experiment_url = "%s/api/v1/experiment/1/?format=json" % config.url
        mocker.get(get_experiment_url, text=mock_get_response)
        experiment = Experiment.objects.get(id=1)
        assert experiment.response_dict == mock_experiment


def test_experiment_create():
    """
    Test creating a experiment record
    """
    mock_experiment = {
        "id": 1,
        "title": "Test Experiment",
        "description": "exp description",
        "authors": [
        ],
        "created_by": "/api/v1/user/1/",
        "created_time": "2017-08-03T12:59:48.600082",
        "institution_name": "Monash University",
        "owner_ids": [
            1,
        ],
        "parameter_sets": [
        ],
        "public_access": 1,
        "update_time": "2017-08-03T12:59:48.600106",
        "resource_uri": "/api/v1/experiment/1/",
    }
    mock_post_response = json.dumps(mock_experiment)
    with requests_mock.Mocker() as mocker:
        post_experiment_url = "%s/api/v1/experiment/" % config.url
        mocker.post(post_experiment_url, text=mock_post_response)
        experiment = Experiment.objects.create(title="experiment title")
        assert experiment.response_dict == mock_experiment



"""
Model class for MyTardis API v1's ReplicaResource.
"""


class Replica(object):
    """
    Model class for MyTardis API v1's ReplicaResource.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, response_dict):
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        if 'location' in response_dict:
            self.location = response_dict['location']
        else:
            self.location = ''
        self.uri = response_dict['uri']
        self.verified = response_dict['verified']



"""
Model class for MyTardis API v1's InstrumentResource.
"""
from __future__ import print_function

import json
import logging

import requests

from ..conf import config
from .facility import Facility
from .model import Model
from .resultset import ResultSet

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Instrument(Model):
    """
    Model class for MyTardis API v1's InstrumentResource.
    """
    def __init__(self, response_dict):
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.name = response_dict['name']
        self.response_dict = response_dict
        self.facility = Facility(response_dict['facility'])

    def __str__(self):
        """
        Return a string representation of an instrument
        """
        return "<%s: %s>" % (type(self).__name__, self.name)

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of instruments

        :param filters: Filters, e.g. "id=123" or "facility__id=45"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`Instrument` records, encapsulated in a
            ResultSet object.
        """
        from ..utils import add_filters, extend_url

        url = "%s/api/v1/instrument/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(Instrument, url, response.json())

    @staticmethod
    def get(**kwargs):
        r"""
        Retrieve a single instrument record

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the Instrument to retrieve

        :return: A :class:`Instrument` record.

        :raises requests.exceptions.HTTPError:
        """
        instrument_id = kwargs.get("id")
        if not instrument_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for Instrument get "
                "at this stage.")
        url = "%s/api/v1/instrument/%s/?format=json" % \
            (config.url, instrument_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return Instrument(response.json())

    @staticmethod
    def create(facility_id, name):
        """
        Create an instrument record.

        :param facility_id: The ID of the facility to create the instrument in.
        :param name: The name of the instrument.

        :return: A new :class:`Instrument` record.
        """
        new_instrument_json = {
            "name": name,
            "facility": "/api/v1/facility/%s/" % facility_id
        }
        url = "%s/api/v1/instrument/" % config.url
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_instrument_json))
        response.raise_for_status()
        return Instrument(response.json())

    @staticmethod
    def update(instrument_id, name):
        """
        Update an instrument record.

        :param instrument_id: The ID of the instrument record to update.
        :param name: The new name of the instrument.

        :return: An updated :class:`Instrument` record.
        """
        updated_fields_json = {
            "name": name,
        }
        url = "%s/api/v1/instrument/%s/" % (config.url, instrument_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        response.raise_for_status()
        return Instrument(response.json())



"""
Base class for models to inherit from
"""
import six
from six import with_metaclass

from .queryset import QuerySet

class Manager(object):
    """
    Each Model.objects instance will be an instance of this
    Manager class.
    """
    @classmethod
    def get(cls, **kwargs):
        """
        Override this with a method to retrieve a single instance of the model
        """

    @classmethod
    def all(cls, **kwargs):
        """
        Override this to retrieve all instances of the model

        Initially this is being implemented for result sets, with the intention
        to also implement it for query sets, for which Model.objects.all() will
        return a generator which will not trigger any API requests until we
        attempt to index it or convert it to a list etc.
        """

    @classmethod
    def filter(cls, **kwargs):
        """
        Override this to provide a method to filter by kwargs.

        Initially this is being implemented for result sets, with the intention
        to also implement it for query sets, for which Model.objects.filter()
        will return a generator which will not trigger any API requests until
        we attempt to index it or convert it to a list etc.
        """

    @classmethod
    def create(cls, **kwargs):
        """
        Override this with a method to create a single instance of the model
        """


class ModelMetaclass(type):
    """
    Metaclass to allow mytardisclient model classes to use
    Django-style syntax (Model.objects.get etc.) to access
    their methods from a .objects property
    """
    @property
    def objects(cls):
        """
        Provides the .objects property, allowing the model's
        static methods to be accessed with Model.objects.[method]
        """
        if not hasattr(cls, "_objects"):
            cls._objects = Manager()

        setattr(cls._objects, "get", getattr(cls, "get"))
        if hasattr(cls, "create"):
            setattr(cls._objects, "create", getattr(cls, "create"))

        def all_method():
            return QuerySet(cls)

        setattr(cls._objects, "all", all_method)

        def filter_method(filters):
            return QuerySet(cls, filters=filters)

        def _filter(**kwargs):
            filter_str = "&".join(
                "%s=%s" % (key, value) for key, value in six.iteritems(kwargs))
            return filter_method(filters=filter_str)

        setattr(cls._objects, "filter", _filter)

        def order_by_method(order_by):
            return QuerySet(cls, order_by=order_by)

        def _order_by(order_by):
            return order_by_method(order_by=order_by)

        setattr(cls._objects, "order_by", _order_by)

        return cls._objects


class Model(with_metaclass(ModelMetaclass, object)):
    """
    Base class for models to inherit from
    """
    # pylint: disable=too-few-public-methods
    def __repr__(self):
        """
        Return a string representation
        """
        if hasattr(self, "__str__"):
            return self.__str__()
        return super(Model, self).__repr__()



"""
Model class for MyTardis API v1's SchemaResource.
"""
from __future__ import print_function

import logging
import re

import requests

from ..conf import config
from ..utils import extend_url, add_filters
from .model import Model
from .resultset import ResultSet

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Schema(Model):
    """
    Model class for MyTardis API v1's SchemaResource.
    """
    # pylint: disable=too-many-instance-attributes
    def __init__(self, response_dict, param_names=False):
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.name = response_dict['name']
        self.hidden = response_dict['hidden']
        self.immutable = response_dict['immutable']
        self.namespace = response_dict['namespace']
        type_index = response_dict['type']
        _schema_types = ['', 'Experiment schema', 'Dataset schema', 'Datafile schema',
                         'None', 'Instrument schema']
        self.type = _schema_types[type_index]  # pylint: disable=invalid-name
        self.subtype = response_dict['subtype']

        if param_names:
            self.parameter_names = ParameterName.list(
                filters="schema__id=%s" % self.id)
        else:
            self.parameter_names = ResultSet.empty(ParameterName)

    def __str__(self):
        """
        Return a string representation of a schema
        """
        return "<%s: %s>" % (type(self).__name__, self.name)

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of schemas.

        :param filters: Filters, e.g. "id=123" or "namespace=NameSpace"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`Schema` records, encapsulated in a
            `ResultSet` object`.
        """
        url = "%s/api/v1/schema/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(Schema, url, response.json())

    @staticmethod
    def get(**kwargs):
        r"""
        Retrieve a single schema record

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the Schema to retrieve

        :return: A :class:`Schema` record.

        :raises requests.exceptions.HTTPError:
        """
        schema_id = kwargs.get("id")
        if not schema_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for Schema get "
                "at this stage.")
        param_names = kwargs.get("param_names", False)
        url = "%s/api/v1/schema/%s/?format=json" % (config.url, schema_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return Schema(response.json(), param_names)


class ParameterName(Model):
    """
    Model class for MyTardis API v1's ParameterNameResource.
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, response_dict):
        self.response_dict = response_dict
        schema_id = response_dict['schema'].split('/')[-2]
        self.schema = Schema.objects.get(id=schema_id)
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.name = response_dict['name']
        self.full_name = response_dict['full_name']
        _type_choices = ['', 'Numeric', 'String', 'URL', 'Link',
                         'Filename', 'DateTime', 'Long String', 'JSON']
        self.data_type = _type_choices[response_dict['data_type']]
        self.units = response_dict['units']
        self.immutable = response_dict['immutable']
        self.is_searchable = response_dict['is_searchable']
        self.order = response_dict['order']
        self.choices = response_dict['choices']
        _comparison_types = \
            ['', 'Exact value', 'Not equal',
             'Range', 'Greater than', 'Greater than or equal to',
             'Less than', 'Less than or equal to', 'Contains']
        self.comparison_type = \
            _comparison_types[response_dict['comparison_type']]

    def __str__(self):
        """
        Return a string representation of a parameter name
        """
        return "<%s: %s>" % (type(self).__name__, self.full_name)

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of parameter names.

        :param filters: Filters, e.g. "schema__id=123" or "name=ParamName"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`ParameterName` records,
            encapsulated in a `ResultSet` object`.
        """
        url = "%s/api/v1/parametername/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        parameter_names_dict = response.json()
        num_records = len(parameter_names_dict['objects'])

        schema_id = None
        filter_components = filters.split("&")
        for filter_component in filter_components:
            match = re.search(r"schema__id=([0-9]+)", filter_component)
            if match:
                schema_id = match.groups()[0]
        if not schema_id:
            return ResultSet(ParameterName, url, parameter_names_dict)

        schema_resource_uri = "/api/v1/schema/%s/" % schema_id
        parameter_names_dict['objects'] = \
            [pn for pn in parameter_names_dict['objects']
             if pn['schema'] == schema_resource_uri]

        offset = 0
        limit = parameter_names_dict['meta']['limit']
        total_count = parameter_names_dict['meta']['total_count']
        while num_records < total_count:
            offset += limit
            url = "%s/api/v1/parametername/?format=json" % config.url
            url += "&offset=%s" % offset
            response = requests.get(url=url, headers=config.default_headers)
            response.raise_for_status()
            parameter_names_page_dict = response.json()
            num_records += len(parameter_names_page_dict['objects'])
            parameter_names_page_dict['objects'] = \
                [pn for pn in parameter_names_page_dict['objects']
                 if pn['schema'] == schema_resource_uri]
            parameter_names_dict['objects'].extend(parameter_names_page_dict['objects'])

        return ResultSet(ParameterName, url, parameter_names_dict)

    @staticmethod
    def get(**kwargs):
        r"""
        Retrieve a single parameter name record

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the ParameterName to retrieve

        :return: A :class:`ParameterName` record.
        """
        pname_id = kwargs.get("id")
        if not pname_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for ParameterName "
                "get at this stage.")
        url = "%s/api/v1/parametername/%s/?format=json" % (config.url,
                                                           pname_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ParameterName(response.json())



"""
test_api_endpoints.py

Tests for functionality to list the available API endpoints of a remote
MyTardis server
"""
import json
import requests_mock

from mtclient.conf import config
from mtclient.models.api import ApiEndpoint, ApiSchema


def test_api_list():
    """
    Test listing the API endpoints of a remote MyTardis server
    """
    mock_api_endpoints = {
        "dataset": {
            "list_endpoint": "/api/v1/dataset/",
            "schema": "/api/v1/dataset/schema/"
        },
        "dataset_file": {
            "list_endpoint": "/api/v1/dataset_file/",
            "schema": "/api/v1/dataset_file/schema/"
        },
        "experiment": {
            "list_endpoint": "/api/v1/experiment/",
            "schema": "/api/v1/experiment/schema/"
        }
    }
    mock_api_list_response = json.dumps(mock_api_endpoints)
    with requests_mock.Mocker() as mocker:
        list_api_endpoints_url = "%s/api/v1/?format=json" % config.url
        mocker.get(list_api_endpoints_url, text=mock_api_list_response)
        endpoints = ApiEndpoint.list()
        returned_endpoints = \
            {endpoint.model: endpoint.response_dict for endpoint in endpoints}
        assert returned_endpoints == mock_api_endpoints


def test_api_get_schema():
    """
    Test listing the API endpoints of a remote MyTardis server
    """
    mock_api_schema = {
        "allowed_detail_http_methods": [
            "get",
            "post"
        ],
        "allowed_list_http_methods": [
            "get",
            "post"
        ],
        "default_format": "application/json",
        "default_limit": 20,
        "fields": {
            "description": {
                "blank": True,
                "default": "",
                "nullable": False,
                "primary_key": False,
                "readonly": False,
                "type": "string",
                "unique": False,
                "verbose_name": "description"
            }
        },
        "filtering": {
            "description": [
                "exact"
            ]
        },
        "ordering": [
            "description"
        ]
    }
    mock_api_response = json.dumps(mock_api_schema)
    with requests_mock.Mocker() as mocker:
        url = "%s/api/v1/dataset/schema/?format=json" % config.url
        mocker.get(url, text=mock_api_response)
        schema = ApiSchema.get("dataset")
        assert schema.response_dict == mock_api_schema



"""
setup.py for mytardisclient

Build docs:
  python setup.py build_sphinx

Upload source distribution:
    # Put credentials in ~/.pypirc
    rm dist/*
    python setup.py sdist
    pip install twine
    twine upload dist/*
"""
import os
from setuptools import setup
from setuptools import find_packages

from mtclient.version import VERSION

# read the contents of your README file
THIS_DIR = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(THIS_DIR, 'README.rst')) as readme:
    LONG_DESCRIPTION = readme.read()

setup(name='mytardisclient',
      packages=find_packages(),
      version=VERSION,
      description="Command Line Interface and Python classes "
      "for interacting with MyTardis's REST API.",
      long_description=LONG_DESCRIPTION,
      long_description_content_type='text/x-rst',
      author='James Wettenhall',
      author_email='james.wettenhall@monash.edu',
      url='http://github.com/mytardis/mytardisclient',
      download_url='https://github.com/mytardis/mytardisclient'
      '/archive/%s.tar.gz' % VERSION,
      keywords=['mytardis', 'REST'], # arbitrary keywords
      classifiers=[],
      license='GPL',
      entry_points={
          "console_scripts": [
              "mytardis = mtclient.client:run",
          ],
      },
      install_requires=['requests', 'pyopenssl', 'ndg-httpsclient', 'pyasn1',
                        'configparser', 'texttable', 'dogpile.cache', 'six',
                        'clint'],
      zip_safe=False)



"""
test_facility.py

Tests for functionality to query the Facility model via
the MyTardis REST API's facility resource
"""
import json
import requests_mock

from mtclient.conf import config
from mtclient.models.facility import Facility


def test_facility_list():
    """
    Test listing facility records exposed in a MyTardis REST API
    """
    mock_facility_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [
            {
                "created_time": None,
                "id": 1,
                "manager_group": {
                    "id": 1,
                    "name": "test-facility-managers",
                    "resource_uri": "/api/v1/group/1/"
                },
                "modified_time": None,
                "name": "Test Facility",
                "resource_uri": "/api/v1/facility/1/"
            }
        ]
    }
    mock_list_response = json.dumps(mock_facility_list)
    with requests_mock.Mocker() as mocker:
        list_facilities_url = "%s/api/v1/facility/?format=json" % config.url
        mocker.get(list_facilities_url, text=mock_list_response)
        facilities = Facility.objects.all()
        facilities._execute_query()
        assert facilities._result_set.response_dict == mock_facility_list


def test_facility_get():
    """
    Test getting a facility record by ID
    """
    mock_facility = {
        "created_time": None,
        "id": 1,
        "manager_group": {
            "id": 1,
            "name": "test-facility-managers",
            "resource_uri": "/api/v1/group/1/"
        },
        "modified_time": None,
        "name": "Test Facility",
        "resource_uri": "/api/v1/facility/1/"
    }
    mock_get_response = json.dumps(mock_facility)
    with requests_mock.Mocker() as mocker:
        get_facility_url = "%s/api/v1/facility/1/?format=json" % config.url
        mocker.get(get_facility_url, text=mock_get_response)
        facility = Facility.objects.get(id=1)
        assert facility.response_dict == mock_facility



"""
Model class for MyTardis API v1's DatasetResource.
"""
from __future__ import print_function

import logging

import requests
from six.moves import urllib

from ..conf import config
from ..utils import extend_url, add_filters

from .resultset import ResultSet
from .schema import Schema
from .schema import ParameterName
from .instrument import Instrument
from .model import Model

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name


class Dataset(Model):
    """
    Model class for MyTardis API v1's DatasetResource.
    """
    def __init__(self, response_dict=None, include_metadata=False):
        self.response_dict = response_dict
        self.id = None  # pylint: disable=invalid-name
        self.description = None
        self.instrument = None
        self.experiments = []
        for key in self.__dict__:
            if key in response_dict:
                self.__dict__[key] = response_dict[key]
        if response_dict['instrument']:
            self.instrument = Instrument(response_dict['instrument'])
        self.parameter_sets = []
        if include_metadata:
            for dataset_param_set_json in response_dict['parameter_sets']:
                self.parameter_sets.append(
                    DatasetParameterSet(dataset_param_set_json))

    def __str__(self):
        """
        Return a string representation of a dataset
        """
        return "<%s: %s>" % (type(self).__name__, self.description)

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of datasets.

        :param filters: Filters, e.g. "experiments__id=123&description=Dataset Description"
        :param limit: Maximum number of results to return.
        :param offset: Skip this many records from the start of the result set.
        :param order_by: Order by this field.

        :return: A list of :class:`Dataset` records, encapsulated in a
            `ResultSet` object.
        """
        url = "%s/api/v1/dataset/?format=json" % config.url

        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(Dataset, url, response.json())

    @staticmethod
    def get(**kwargs):
        r"""
        Retrieve a single dataset record

        :param \**kwargs:
          See below

        :Keyword Arguments:
            * *id* (``int``) --
              ID of the Dataset to retrieve

        :return: A :class:`Dataset` record.

        :raises requests.exceptions.HTTPError:
        """
        dataset_id = kwargs.get("id")
        if not dataset_id:
            raise NotImplementedError(
                "Only the id keyword argument is supported for Dataset get "
                "at this stage.")
        include_metadata = kwargs.get("include_metadata", False)
        url = "%s/api/v1/dataset/%s/?format=json" % (config.url, dataset_id)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return Dataset(response_dict=response.json(),
                       include_metadata=include_metadata)

    @staticmethod
    def create(experiment_id, description, instrument_id=None,
               params_file_json=None):
        """
        Create a dataset record.

        :param experiment_id: The ID of the experiment to create the
            dataset in.
        :param description: The description of the dataset.
        :param instrument_id: The instrument the data was collected on.
        :param params_file_json: Path to a JSON file with dataset
            parameters.

        :return: A new :class:`Dataset` record.
        """
        import json
        import os

        new_dataset_json = {
            "description": description,
            "experiments": ["/api/v1/experiment/%s/" % experiment_id],
            "immutable": False
        }
        if instrument_id:
            new_dataset_json['instrument'] = "/api/v1/instrument/%s/" \
                % instrument_id
        if params_file_json:
            assert os.path.exists(params_file_json)
            with open(params_file_json) as params_file:
                new_dataset_json['parameter_sets'] = json.load(params_file)
        url = "%s/api/v1/dataset/" % config.url
        response = requests.post(headers=config.default_headers, url=url,
                                 data=json.dumps(new_dataset_json))
        response.raise_for_status()
        return Dataset(response.json())

    @staticmethod
    def update(dataset_id, description):
        """
        Update an dataset record.
        """
        import json

        updated_fields_json = {'description': description}
        url = "%s/api/v1/dataset/%s/" % (config.url, dataset_id)
        response = requests.patch(headers=config.default_headers, url=url,
                                  data=json.dumps(updated_fields_json))
        response.raise_for_status()
        dataset_json = response.json()
        return Dataset(dataset_json)

    @staticmethod
    def download(dataset_id):
        """
        Download a dataset
        """
        import os
        from .datafile import DataFile

        dataset = Dataset.objects.get(id=dataset_id)
        path = urllib.parse.quote(
            dataset.description.encode('utf-8'), safe=" ,")
        if os.path.exists(path):
            from ..utils.confirmation import query_yes_no
            if not query_yes_no("Overwrite '%s/'?" % path):
                return
        else:
            os.makedirs(path)
        print("Downloading to: %s/" % path)
        for datafile in DataFile.objects.filter(dataset__id=dataset_id).order_by('id'):
            DataFile.download(datafile.id, basedir=path, overwrite=True)
        print("Downloaded to: %s/" % path)


class DatasetParameterSet(object):
    """
    Model class for MyTardis API v1's DatasetParameterSetResource.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, response_dict):
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        self.dataset = response_dict['dataset']
        self.schema = Schema(response_dict['schema'])
        self.parameters = []
        for dataset_param_json in response_dict['parameters']:
            self.parameters.append(DatasetParameter(dataset_param_json))

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        Retrieve a list of dataset parameters.

        :param filters: Filters, e.g. "datasets__id=123"

        :return: A list of :class:`DatasetParameterSet` records,
            encapsulated in a `ResultSet` object`.
        """
        url = "%s/api/v1/datasetparameterset/?format=json" % config.url
        url = add_filters(url, filters)
        url = extend_url(url, limit, offset, order_by)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ResultSet(DatasetParameterSet, url, response.json())


class DatasetParameter(object):
    """
    Model class for MyTardis API v1's DatasetParameterResource.
    """
    # pylint: disable=too-few-public-methods
    # pylint: disable=too-many-instance-attributes
    def __init__(self, response_dict):
        self.response_dict = response_dict
        self.id = response_dict['id']  # pylint: disable=invalid-name
        pname_id = response_dict['name'].split('/')[-2]
        self.name = ParameterName.objects.get(id=pname_id)
        self.string_value = response_dict['string_value']
        self.numerical_value = response_dict['numerical_value']
        self.datetime_value = response_dict['datetime_value']
        self.link_id = response_dict['link_id']
        self.value = response_dict['value']

    @staticmethod
    def list(filters=None, limit=None, offset=None, order_by=None):
        """
        List dataset parameter records in parameter set.
        """



"""
The api module contains model classes for MyTardis API v1's endpoints.

This following lists all of the supported endpoints: /api/v1/?format=json

API functionality available for a particular model can be retrieved with:
    /api/v1/facility/schema/?format=json

The 'schema' request above requires authentication.
"""
from __future__ import print_function

import requests
import six

from ..conf import config


class ApiEndpoint(object):
    """
    Model class for MyTardis API v1's endpoints.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, model, endpoint_dict):
        """
        :param model: The name of the API model resource, e.g. 'dataset_file'
        :param endpoint_dict: A dictionary with two keys: 'list_endpoint' and
            'schema', e.g. {"list_endpoint": "/api/v1/dataset_file/",
            "schema": "/api/v1/dataset_file/schema/"}. This dictionary is
            obtained by isolating a single model resource from the
            deserialized response from /api/v1/?format=json
        """
        self.response_dict = endpoint_dict
        self.model = model
        self.list_endpoint = endpoint_dict['list_endpoint']
        self.schema = endpoint_dict['schema']

    def __str__(self):
        """
        Return a string representation of an API endpoint
        """
        return "%s: %s, %s" % (self.model, self.list_endpoint, self.schema)

    def __repr__(self):
        """
        Return a string representation of an API endpoint
        """
        return self.__str__()

    @staticmethod
    def list():
        """
        Retrieve a list of API endpoints, encapsulated in
        an :class:`ApiEndpoints` object.

        The :class:`ApiEndpoints` object encapsulates the entire JSON response
        from the /api/v1/ query.

        :return: A list of API endpoints, encapsulated in
            an :class:`ApiEndpoints` object.
        """
        url = "%s/api/v1/?format=json" % config.url
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ApiEndpoints(response.json())


class ApiSchema(object):
    """
    Model class for MyTardis API v1's schemas.
    """
    # pylint: disable=too-many-instance-attributes
    # pylint: disable=too-few-public-methods
    def __init__(self, model, response_dict):
        """
        :param model: The name of an API-accessible model, e.g. 'dataset_file'.
        :param response_dict: The JSON-deserialized response from an
            /api/v1/model/schema/ query
        """
        self.model = model
        self.response_dict = response_dict
        self.fields = response_dict['fields']
        self.filtering = response_dict['filtering'] if 'filtering' in response_dict else {}
        for key, val in six.iteritems(self.filtering):
            if val == 1:
                self.filtering[key] = "ALL"
            elif val == 2:
                self.filtering[key] = "ALL_WITH_RELATIONS"
        self.ordering = response_dict['ordering'] if 'ordering' in response_dict else {}
        self.allowed_list_http_methods = response_dict['allowed_list_http_methods']
        self.allowed_detail_http_methods = response_dict['allowed_detail_http_methods']
        self.default_format = response_dict['default_format']
        self.default_limit = response_dict['default_limit']

    @staticmethod
    def get(model):
        """
        Get a list of API-accessible functionality for a particular model.

        :param model: The name of an API-accessible model, e.g. 'dataset_file'.
        :return: An :class:`ApiSchema` object, which encapsulates the list of
                API-accessible functionality for a particular model.
        """
        if model == "datafile":
            model = "dataset_file"
        url = "%s/api/v1/%s/schema/?format=json" % (config.url, model)
        response = requests.get(url=url, headers=config.default_headers)
        response.raise_for_status()
        return ApiSchema(model, response.json())


class ApiEndpoints(object):
    """
    Dictionary of API endpoints (list_endpoint and schema)
    with model names as keys.
    """
    # pylint: disable=too-few-public-methods
    def __init__(self, response_dict):
        """
        Dictionary of API endpoints with model names as keys.

        :param response_dict: The deserialized response from the /api/v1/ query
        """
        self.response_dict = response_dict
        self.total_count = len(self.response_dict.keys())

    def __len__(self):
        """
        Return the number of models accessible via the API.
        :return: The number of models accessible via the API.
        """
        return len(self.response_dict.keys())

    def __getitem__(self, model):
        """
        Return the API endpoint for a particular model.

        :param model: The name of an API-accessible model, e.g. 'dataset_file'.
        :return: The :class:`ApiEndpoint` object for that supplied model.
        """
        return ApiEndpoint(model, self.response_dict[model])

    def __iter__(self):
        """
        Iterate the :class:`ApiEndpoints` set.
        """
        for index in range(0, len(self)):
            model = sorted(list(self.response_dict.keys()))[index]
            yield ApiEndpoint(model, self.response_dict[model])



"""
test_datafile.py

Tests for functionality to query the DataFile model via
the MyTardis REST API's datafile resource
"""
import json
import tempfile

import pytest
import requests_mock

from mtclient.conf import config
from mtclient.models.datafile import DataFile


def test_datafile_list():
    """
    Test listing datafile records exposed in a MyTardis REST API
    """
    mock_datafile_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 1
        },
        "objects": [{
            "id": 1,
            "created_time": "2016-11-10T13:50:25.258483",
            "dataset": "/api/v1/dataset/1/",
            "directory": "subdir",
            "filename": "testfile1.txt",
            "md5sum": "bogus",
            "mimetype": "text/plain",
            "modification_time": None,
            "parameter_sets": [
            ],
            "replicas": [
                {
                    "datafile": "/api/v1/dataset_file/1/",
                    "id": 1,
                    "location": "local box at /home/mytardis/var/local",
                    "resource_uri": "/api/v1/replica/1/",
                    "uri": "subdir/testfile1.txt",
                    "verified": True
                }
            ],
            "resource_uri": "/api/v1/dataset_file/1/",
            "size": 32,
        }]
    }
    mock_list_response = json.dumps(mock_datafile_list)
    with requests_mock.Mocker() as mocker:
        list_datafiles_url = "%s/api/v1/dataset_file/?format=json&dataset__id=1" % config.url
        mocker.get(list_datafiles_url, text=mock_list_response)
        datafiles = DataFile.list(filters="dataset__id=1")
        assert datafiles.response_dict == mock_datafile_list


def test_datafile_get():
    """
    Test getting a datafile record by ID
    """
    mock_datafile = {
        "id": 1,
        "created_time": "2016-11-10T13:50:25.258483",
        "dataset": "/api/v1/dataset/1/",
        "directory": "subdir",
        "filename": "testfile1.txt",
        "md5sum": "bogus",
        "mimetype": "text/plain",
        "modification_time": None,
        "parameter_sets": [
        ],
        "replicas": [
            {
                "datafile": "/api/v1/dataset_file/1/",
                "id": 1,
                "location": "local box at /home/mytardis/var/local",
                "resource_uri": "/api/v1/replica/1/",
                "uri": "subdir/testfile1.txt",
                "verified": True
            }
        ],
        "resource_uri": "/api/v1/dataset_file/1/",
        "size": 32,
    }
    mock_get_response = json.dumps(mock_datafile)
    with requests_mock.Mocker() as mocker:
        get_datafile_url = "%s/api/v1/dataset_file/1/?format=json" % config.url
        mocker.get(get_datafile_url, text=mock_get_response)
        datafile = DataFile.objects.get(id=1)
        assert datafile.response_dict == mock_datafile
        assert str(datafile) == "<DataFile: subdir/testfile1.txt>"
        with pytest.raises(NotImplementedError) as err:
            DataFile.objects.get(invalid_key="value")
            assert str(err) == (
                "Only the id keyword argument is supported for DataFile get "
                "at this stage.")


def test_datafile_create():
    """
    Test creating a datafile record
    """
    # DataFile.create_datafile(...) will look up the Dataset specified by
    # its dataset_id argument, so we need to mock the response for this
    # Dataset lookup:
    mock_dataset = {
        "created_time": None,
        "description": "dataset description",
        "directory": None,
        "experiments": [
            "/api/v1/experiment/1/"
        ],
        "id": 1,
        "immutable": False,
        "instrument": None,
        "modified_time": None,
        "parameter_sets": [],
        "resource_uri": "/api/v1/dataset/1/"
    }
    mock_ds_get_response = json.dumps(mock_dataset)

    # DataFile.create_datafile(...) will check whether a DataFile already
    # exists with this dataset ID, filename and directory, so we need to
    # mock the response for this DataFile lookup:
    mock_datafile_list = {
        "meta": {
            "limit": 20,
            "next": None,
            "offset": 0,
            "previous": None,
            "total_count": 0
        },
        "objects": [
        ]
    }
    mock_df_list_response = json.dumps(mock_datafile_list)

    # The MyTardis API's DataFile creation endpoint returns the newly
    # created DataFile ID in the response header, but DataFile.create_datafile
    # needs to query the API again to get the full details of the newly created
    # DataFile record, so we'll mock this DataFile lookup:
    mock_new_datafile = {
        "id": 1,
        "created_time": "2016-11-10T13:50:25.258483",
        "dataset": "/api/v1/dataset/1/",
        "directory": "subdir",
        "filename": "testfile1.txt",
        "md5sum": "bogus",
        "mimetype": "text/plain",
        "modification_time": None,
        "parameter_sets": [
        ],
        "replicas": [
            {
                "datafile": "/api/v1/dataset_file/1/",
                "id": 1,
                "location": "local box at /home/mytardis/var/local",
                "resource_uri": "/api/v1/replica/1/",
                "uri": "subdir/testfile1.txt",
                "verified": True
            }
        ],
        "resource_uri": "/api/v1/dataset_file/1/",
        "size": 32,
    }
    mock_new_df_response = json.dumps(mock_new_datafile)
    with requests_mock.Mocker() as mocker:
        get_dataset_url = "%s/api/v1/dataset/1/?format=json" % config.url
        mocker.get(get_dataset_url, text=mock_ds_get_response)
        df_list_url = ("%s/api/v1/dataset_file/?format=json"
                       "&filename=testfile.txt"
                       "&directory=subdir" % config.url)
        mocker.get(df_list_url, text=mock_df_list_response)
        post_datafile_url = "%s/api/v1/dataset_file/" % config.url
        mocker.post(post_datafile_url,
                    headers=dict(location="/api/v1/dataset_file/1/"))
        get_new_df_url = "%s/api/v1/dataset_file/1/?format=json" % config.url
        mocker.get(get_new_df_url, text=mock_new_df_response)
        datafile = DataFile.create_datafile(
            dataset_id=1,
            storagebox="local box at /home/mytardis/var/local",
            dataset_path="/path/to/dataset1",
            file_path="/path/to/dataset1/subdir/testfile.txt",
            check_local_paths=False,
            create_dataset_symlink=False,
            size=32,
            md5sum="bogus")
        assert datafile.response_dict == mock_new_datafile


def test_md5_sum():
    """
    Test calculating the MD5 sum of a datafile
    """
    import os
    import sys

    from mtclient.models.datafile import md5_sum

    with tempfile.NamedTemporaryFile() as tmpfile:
        tmpfile_name = tmpfile.name

    with open(tmpfile_name, 'w') as tmpfile:
        tmpfile.write("Hello, world!\n")

    assert md5_sum(tmpfile_name) == "746308829575e17c3331bbcb00c0898b"

    try:
        os.remove(tmpfile_name)
    except IOError as err:
        sys.stderr.write("%s\n" % err)



"""
Views for MyTardis instrument records
"""
from __future__ import print_function

import json
from texttable import Texttable


def render_instrument(instrument, render_format):
    """
    Render instrument

    :param instrument: The instrument to be rendered.
    :type instrument: :class:`mtclient.models.instrument.Instrument`
    :param render_format: The format to display the data in ('table' or
        'json').
    """
    if render_format == 'json':
        return render_instrument_as_json(instrument)
    return render_instrument_as_table(instrument)


def render_instrument_as_json(instrument, indent=2, sort_keys=True):
    """
    Returns JSON representation of instrument.

    :param instrument: The instrument to be rendered.
    :type instrument: :class:`mtclient.models.instrument.Instrument`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        instrument.response_dict, indent=indent, sort_keys=sort_keys)


def render_instrument_as_table(instrument):
    """
    Returns ASCII table view of instrument.

    :param instrument: The instrument to be rendered.
    :type instrument: :class:`mtclient.models.instrument.Instrument`
    """
    instrument_table = Texttable()
    instrument_table.set_cols_align(['l', 'l'])
    instrument_table.set_cols_valign(['m', 'm'])
    instrument_table.header(["Instrument field", "Value"])
    instrument_table.add_row(["ID", instrument.id])
    instrument_table.add_row(["Name", instrument.name])
    instrument_table.add_row(["Facility", instrument.facility])
    return instrument_table.draw() + "\n"


def render_instruments(instruments, render_format, display_heading=True):
    """
    Render instruments

    :param instruments: The `ResultSet` of instruments to be rendered.
    :type instruments: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: When using the 'table' render format for a
        `ResultSet` containing multiple records, setting
        `display_heading` to True ensures that the meta information
        returned by the query is summarized in a 'heading' before
        displaying the table.  This meta information can be used to
        determine whether the query results have been truncated due
        to pagination.
    """
    if render_format == 'json':
        return render_instruments_as_json(instruments)
    return render_instruments_as_table(instruments, display_heading)


def render_instruments_as_json(instruments, indent=2, sort_keys=True):
    """
    Returns JSON representation of instruments.

    :param instruments: The result set of instruments to be displayed.
    :type instruments: :class:`mtclient.models.resultset.ResultSet`
    :param indent: If indent is a non-negative integer or string, then JSON
        array elements and object members will be pretty-printed with that
        indent level.
    :param sort_keys: If sort_keys is `True` (default: `False`), then the
        rendered JSON will be sorted by key.
    """
    return json.dumps(
        instruments.response_dict, indent=indent, sort_keys=sort_keys)


def render_instruments_as_table(instruments, display_heading=True):
    """
    Returns ASCII table view of instruments.

    :param instruments: The instruments to be rendered.
    :type instruments: :class:`mtclient.models.resultset.ResultSet`
    :param render_format: The format to display the data in ('table' or
        'json').
    :param display_heading: Setting `display_heading` to True ensures
        that the meta information returned by the query is summarized
        in a 'heading' before displaying the table.  This meta
        information can be used to determine whether the query results
        have been truncated due to pagination.
    """
    heading = "\n" \
        "Model: Instrument\n" \
        "Query: %s\n" \
        "Total Count: %s\n" \
        "Limit: %s\n" \
        "Offset: %s\n\n" \
        % (instruments.url, instruments.total_count,
           instruments.limit, instruments.offset) if display_heading else ""

    table = Texttable(max_width=0)
    table.set_cols_align(["r", 'l', 'l'])
    table.set_cols_valign(['m', 'm', 'm'])
    table.header(["ID", "Name", "Facility"])
    for instrument in instruments:
        table.add_row([instrument.id, instrument.name, instrument.facility])
    return heading + table.draw() + "\n"



"""
test_api_cli.py

Tests for querying the REST API, listing its endpoints and describing each
endpoint's schema from a command-line interface
"""
import json
import sys
import textwrap
from argparse import Namespace

import requests_mock

import mtclient.client
from mtclient.conf import config
from mtclient.controllers.api import ApiController


def test_api_list_cli_json(capfd):
    """
    Test listing API endpoints, requesting output in JSON format
    """
    mock_api_endpoints = {
        "dataset": {
            "list_endpoint": "/api/v1/dataset/",
            "schema": "/api/v1/dataset/schema/"
        },
        "dataset_file": {
            "list_endpoint": "/api/v1/dataset_file/",
            "schema": "/api/v1/dataset_file/schema/"
        },
        "experiment": {
            "list_endpoint": "/api/v1/experiment/",
            "schema": "/api/v1/experiment/schema/"
        }
    }
    mock_api_list_response = json.dumps(mock_api_endpoints)
    with requests_mock.Mocker() as mocker:
        list_api_endpoints_url = "%s/api/v1/?format=json" % config.url
        mocker.get(list_api_endpoints_url, text=mock_api_list_response)
        api_controller = ApiController()
        args = Namespace(model='api', command='list', json=True, verbose=False)

        api_controller.list(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_api_endpoints

        api_controller.run_command(args)
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_api_endpoints

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'api', 'list', '--json']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_api_endpoints
        sys.argv = sys_argv


def test_api_list_cli_table(capfd):
    """
    Test listing API endpoints, requesting output in ASCII table format
    """
    mock_api_endpoints = {
        "dataset": {
            "list_endpoint": "/api/v1/dataset/",
            "schema": "/api/v1/dataset/schema/"
        },
        "dataset_file": {
            "list_endpoint": "/api/v1/dataset_file/",
            "schema": "/api/v1/dataset_file/schema/"
        },
        "experiment": {
            "list_endpoint": "/api/v1/experiment/",
            "schema": "/api/v1/experiment/schema/"
        }
    }
    mock_api_list_response = json.dumps(mock_api_endpoints)
    expected = textwrap.dedent("""
        API Endpoints
        +--------------+-----------------------+------------------------------+
        |    Model     |     List Endpoint     |            Schema            |
        +==============+=======================+==============================+
        | dataset      | /api/v1/dataset/      | /api/v1/dataset/schema/      |
        +--------------+-----------------------+------------------------------+
        | dataset_file | /api/v1/dataset_file/ | /api/v1/dataset_file/schema/ |
        +--------------+-----------------------+------------------------------+
        | experiment   | /api/v1/experiment/   | /api/v1/experiment/schema/   |
        +--------------+-----------------------+------------------------------+
    """)
    with requests_mock.Mocker() as mocker:
        list_api_endpoints_url = "%s/api/v1/?format=json" % config.url
        mocker.get(list_api_endpoints_url, text=mock_api_list_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'api', 'list']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv


def test_api_get_cli_json(capfd):
    """
    Test getting API endpoint schema via the command-line interface
    """
    mock_api_schema = {
        "allowed_detail_http_methods": [
            "get",
            "post"
        ],
        "allowed_list_http_methods": [
            "get",
            "post"
        ],
        "default_format": "application/json",
        "default_limit": 20,
        "fields": {
            "description": {
                "blank": True,
                "default": "",
                "nullable": False,
                "primary_key": False,
                "readonly": False,
                "type": "string",
                "unique": False,
                "verbose_name": "description"
            }
        },
        "filtering": {
            "description": [
                "exact"
            ]
        },
        "ordering": [
            "description"
        ]
    }
    mock_api_get_response = json.dumps(mock_api_schema)
    with requests_mock.Mocker() as mocker:
        get_api_schema_url = "%s/api/v1/dataset/schema/?format=json" % config.url
        mocker.get(get_api_schema_url, text=mock_api_get_response)
        api_controller = ApiController()
        args = Namespace(
            model='api', api_model='dataset', command='get', json=True,
            verbose=False)
        api_controller.get(args, render_format="json")
        out, _ = capfd.readouterr()
        assert json.loads(out) == mock_api_schema


def test_api_get_cli_table(capfd):
    """
    Test getting API endpoint schema via the command-line interface,
    requesting output in ASCII table format
    """
    mock_api_schema = {
        "allowed_detail_http_methods": [
            "get",
            "post"
        ],
        "allowed_list_http_methods": [
            "get",
            "post"
        ],
        "default_format": "application/json",
        "default_limit": 20,
        "fields": {
            "description": {
                "blank": True,
                "default": "",
                "nullable": False,
                "primary_key": False,
                "readonly": False,
                "type": "string",
                "unique": False,
                "verbose_name": "description"
            }
        },
        "filtering": {
            "description": [
                "exact"
            ]
        },
        "ordering": [
            "description"
        ]
    }
    mock_api_get_response = json.dumps(mock_api_schema)
    expected = textwrap.dedent("""
        +------------------+--------------------+
        | API Schema field |       Value        |
        +==================+====================+
        | Model            | dataset            |
        +------------------+--------------------+
        | Fields           | description        |
        +------------------+--------------------+
        | Filtering        | {                  |
        |                  |   "description": [ |
        |                  |     "exact"        |
        |                  |   ]                |
        |                  | }                  |
        +------------------+--------------------+
        | Ordering         | [                  |
        |                  |   "description"    |
        |                  | ]                  |
        +------------------+--------------------+
    """)
    with requests_mock.Mocker() as mocker:
        get_api_schema_url = "%s/api/v1/dataset/schema/?format=json" % config.url
        mocker.get(get_api_schema_url, text=mock_api_get_response)

        sys_argv = sys.argv
        sys.argv = ['mytardis', 'api', 'get', 'dataset']
        mtclient.client.run()
        out, _ = capfd.readouterr()
        assert out.strip() == expected.strip()
        sys.argv = sys_argv
